/**
 *	Copyright (c) 2008,  Eckhart Arnold
 *
 * Dieses Programm ist freie Software. Sie können es unter den
 * Bedingungen der GNU General Public License, wie von der Free
 * Software Foundation veröffentlicht, weitergeben und/oder
 * modifizieren, entweder gemäß Version 3 der Lizenz oder (nach
 * Ihrer Option) jeder späteren Version.
 *
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung,
 * daß es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE
 * GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder
 * der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden
 * Sie in der GNU General Public License.
 *
 * Sie sollten ein Exemplar der GNU General Public License zusammen
 * mit diesem Programm erhalten haben. Falls nicht, siehe
 * <http://www.gnu.org/licenses/>.
 */

package evolution.examples;

public class PointGenome implements IGenome {
	private TSPData data;
	private static final EAUtil util = new EAUtil();
	private int[] path;
	double stretch = -1.0;
	
	PointGenome(TSPData data) {
		this.data = data;
		EAUtil.RandomIndices rndIndices = util.new RandomIndices(data.points);
		path = new int[data.points];			
		for (int i = 0; i < data.points; i++) {
			path[i] = rndIndices.next();
		}
	}

	PointGenome(PointGenome copy) {
		this.data = copy.data;
		this.path = copy.path; // .clone();
		this.stretch = copy.stretch;
	}
	@Override
	public synchronized IGenome clone() {
		return new PointGenome(this);
	}
	
	public double roundTrip() {
		double x1, y1, x2, y2, dx, dy;
		if (stretch >= 0.0) return stretch;
		else stretch = 0.0;
		x1 = data.location[path[path.length-1]][0];
		y1 = data.location[path[path.length-1]][1];
		for (int i = 0; i < path.length; i++) {
			x2 = data.location[path[i]][0];
			y2 = data.location[path[i]][1];
			dx = x2-x1;  
			dy = y2-y1;
			stretch += Math.sqrt(dx*dx + dy*dy);
			x1 = x2;  
			y1 = y2;
		}
		return stretch;
	}
	
	protected void swap(int a, int b) {
		if (stretch >= 0.0) path = path.clone(); // copy on write !
		
		a %= path.length;
		b %= path.length;
		if (a < 0) a = path.length+a;
		if (b < 0) b = path.length+b;
		int exchg = path[a]; 
		path[a] = path[b]; 
		path[b] = exchg;
		stretch = -1.0;			// invalidate
	}
	
	public void mutate(int a) {
		int b;
		
		// arbitrary mutation algorithm (very bad):
		b = EAUtil.rnd.nextInt(path.length-1);
		if (a == b) b = path.length-1;
	
		// next neighbour exchange mutation algorithm
		// b = a+1; if (b >= path.length) b = 0;
	
		// close neighbour exchange mutation algorithm
		// b = (int) (EAUtil.rnd.nextGaussian() * 2 + a + 0.5);
	
		swap(a, b);	
	}
	
	public void exchange(IGenome other, int from, int to) {
		if (stretch >= 0.0) path = path.clone();  // copy on write 
		if (((PointGenome) other).stretch >= 0.0) 
			((PointGenome) other).path = ((PointGenome) other).path.clone(); 
		
		// two point crossover
		for (int i = from; i < to; i++) {
			int swap = path[i];
			path[i] = ((PointGenome) other).path[i];
			((PointGenome) other).path[i] = swap;
		}
		stretch = -1.0;			// invalidate
		((PointGenome) other).stretch = -1.0;
	}	
	
	protected boolean errorCheck() {
		boolean marker[] = new boolean[path.length];
		for (int i = 0; i < marker.length; i++) marker[i] = false;
		for (int i = 0; i < path.length; i++) {
			if (marker[path[i]]) return false;
			marker[path[i]] = true;
		}
		return true;
	}
	
	public void repair() {
		assert stretch < 0.0;
		
		int fill[] = new int[path.length];
		int stack[][] = new int[path.length][2];
		int sp = 0;
		for (int i = 0; i < path.length; i++) fill[i] = -1;
		for (int i = 0; i < path.length; i++) {
			if (fill[path[i]] < 0) fill[path[i]] = i; 
			else {
				stack[sp][0] = i;
				stack[sp][1] = fill[path[i]];
				sp++;
			}
		}
		for (int i = 0; i < path.length; i++) {
			if (fill[i] < 0) {
				--sp;
				path[stack[sp][EAUtil.rnd.nextInt(2)]] = i; 
			}
		}
		// assert errorCheck();
	}
	
	public int[] getPath() { return path; }
	public int getLength() { return path.length; }
}
